@@ -29,24 +29,22 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: smbfs_subr.c,v 1.9 2003/05/06 21:54:38 lindak Exp $
+ * $Id: smbfs_subr.c,v 1.18 2005/02/02 00:22:23 lindak Exp $
  */
 #include <sys/param.h>
 #include <sys/systm.h>
 #include <sys/kernel.h>
 #include <sys/malloc.h>
-#ifndef APPLE
-#include <machine/clock.h>
-#endif
 #include <sys/time.h>
 #include <sys/vnode.h>
 #include <sys/sysctl.h>
 
-#ifdef APPLE
-#include <sys/smb_apple.h>
-#endif
+#include <sys/kauth.h>
 
-#include <sys/iconv.h>
+#include <sys/smb_apple.h>
+#include <sys/utfconv.h>
+
+#include <sys/smb_iconv.h>
 
 #include <netsmb/smb.h>
 #include <netsmb/smb_conn.h>
@@ -112,14 +110,18 @@ static u_long  lastday;
 static u_short lastddate;
 static u_short lastdtime;
 
-#ifdef APPLE
 PRIVSYM int wall_cmos_clock = 0;	/* XXX */
 PRIVSYM int adjkerntz = 0;	/* XXX */
-#endif
 
 void
-smb_time_local2server(struct timespec *tsp, int tzoff, u_long *seconds)
+smb_time_local2server(struct timespec *tsp, int tzoff, long *seconds)
 {
+	/*
+	 * XXX - what if we connected to the server when it was in
+	 * daylight savings/summer time and we've subsequently switched
+	 * to standard time, or vice versa, so that the time zone
+	 * offset we got from the server is now wrong?
+	 */
 	*seconds = tsp->tv_sec - tzoff * 60 /*- tz.tz_minuteswest * 60 -
 	    (wall_cmos_clock ? adjkerntz : 0)*/;
 }
@@ -127,6 +129,12 @@ smb_time_local2server(struct timespec *tsp, int tzoff, u_long *seconds)
 void
 smb_time_server2local(u_long seconds, int tzoff, struct timespec *tsp)
 {
+	/*
+	 * XXX - what if we connected to the server when it was in
+	 * daylight savings/summer time and we've subsequently switched
+	 * to standard time, or vice versa, so that the time zone
+	 * offset we got from the server is now wrong?
+	 */
 	tsp->tv_sec = seconds + tzoff * 60;
 	    /*+ tz.tz_minuteswest * 60 + (wall_cmos_clock ? adjkerntz : 0)*/;
 }
@@ -134,33 +142,34 @@ smb_time_server2local(u_long seconds, int tzoff, struct timespec *tsp)
 /*
  * Number of seconds between 1970 and 1601 year
  */
-PRIVSYM int64_t DIFF1970TO1601 = 11644473600ULL;
+PRIVSYM u_int64_t DIFF1970TO1601 = 11644473600ULL;
 
 /*
  * Time from server comes as UTC, so no need to use tz
  */
 void
-smb_time_NT2local(int64_t nsec, int tzoff, struct timespec *tsp)
+smb_time_NT2local(u_int64_t nsec, int tzoff, struct timespec *tsp)
 {
 	#pragma unused(tzoff)
 	smb_time_server2local(nsec / 10000000 - DIFF1970TO1601, 0, tsp);
 }
 
 void
-smb_time_local2NT(struct timespec *tsp, int tzoff, int64_t *nsec)
+smb_time_local2NT(struct timespec *tsp, int tzoff, u_int64_t *nsec)
 {
 	#pragma unused(tzoff)
-	u_long seconds;
+	long seconds;
 
 	smb_time_local2server(tsp, 0, &seconds);
-	*nsec = (((int64_t)(seconds) & ~1) + DIFF1970TO1601) * (int64_t)10000000;
+	*nsec = (((u_int64_t)(seconds) & ~1) + DIFF1970TO1601) * (u_int64_t)10000000;
 }
 
 void
 smb_time_unix2dos(struct timespec *tsp, int tzoff, u_int16_t *ddp, 
 	u_int16_t *dtp,	u_int8_t *dhp)
 {
-	u_long t, days, year, month, inc;
+	long t;
+	u_long days, year, month, inc;
 	u_short *months;
 
 	/*
@@ -171,39 +180,58 @@ smb_time_unix2dos(struct timespec *tsp, int tzoff, u_int16_t *ddp,
 	t &= ~1;
 	if (lasttime != t) {
 		lasttime = t;
-		lastdtime = (((t / 2) % 30) << DT_2SECONDS_SHIFT)
-		    + (((t / 60) % 60) << DT_MINUTES_SHIFT)
-		    + (((t / 3600) % 24) << DT_HOURS_SHIFT);
-
-		/*
-		 * If the number of days since 1970 is the same as the last
-		 * time we did the computation then skip all this leap year
-		 * and month stuff.
-		 */
-		days = t / (24 * 60 * 60);
-		if (days != lastday) {
-			lastday = days;
-			for (year = 1970;; year++) {
-				inc = year & 0x03 ? 365 : 366;
-				if (days < inc)
-					break;
-				days -= inc;
-			}
-			months = year & 0x03 ? regyear : leapyear;
-			for (month = 0; days >= months[month]; month++)
-				;
-			if (month > 0)
-				days -= months[month - 1];
-			lastddate = ((days + 1) << DD_DAY_SHIFT)
-			    + ((month + 1) << DD_MONTH_SHIFT);
+		if (t < 0) {
 			/*
-			 * Remember dos's idea of time is relative to 1980.
-			 * unix's is relative to 1970.  If somehow we get a
-			 * time before 1980 then don't give totally crazy
-			 * results.
+			 * This is before 1970, so it's before 1980,
+			 * and can't be represented as a DOS time.
+			 * Just represent it as the DOS epoch.
 			 */
-			if (year > 1980)
-				lastddate += (year - 1980) << DD_YEAR_SHIFT;
+			lastdtime = 0;
+			lastddate = (1 << DD_DAY_SHIFT)
+			    + (1 << DD_MONTH_SHIFT)
+			    + ((1980 - 1980) << DD_YEAR_SHIFT);
+		} else {
+			lastdtime = (((t / 2) % 30) << DT_2SECONDS_SHIFT)
+			    + (((t / 60) % 60) << DT_MINUTES_SHIFT)
+			    + (((t / 3600) % 24) << DT_HOURS_SHIFT);
+
+			/*
+			 * If the number of days since 1970 is the same as
+			 * the last time we did the computation then skip
+			 * all this leap year and month stuff.
+			 */
+			days = t / (24 * 60 * 60);
+			if (days != lastday) {
+				lastday = days;
+				for (year = 1970;; year++) {
+					/*
+					 * XXX - works in 2000, but won't
+					 * work in 2100.
+					 */
+					inc = year & 0x03 ? 365 : 366;
+					if (days < inc)
+						break;
+					days -= inc;
+				}
+				/*
+				 * XXX - works in 2000, but won't work in 2100.
+				 */
+				months = year & 0x03 ? regyear : leapyear;
+				for (month = 0; days >= months[month]; month++)
+					;
+				if (month > 0)
+					days -= months[month - 1];
+				lastddate = ((days + 1) << DD_DAY_SHIFT)
+				    + ((month + 1) << DD_MONTH_SHIFT);
+				/*
+				 * Remember DOS's idea of time is relative
+				 * to 1980, but UN*X's is relative to 1970.
+				 * If somehow we get a time before 1980 then
+				 * don't give totally crazy results.
+				 */
+				if (year > 1980)
+					lastddate += (year - 1980) << DD_YEAR_SHIFT;
+			}
 		}
 	}
 	if (dtp)
@@ -252,6 +280,9 @@ smb_dos2unixtime(u_int dd, u_int dt, u_int dh, int tzoff,
 		year = (dd & DD_YEAR_MASK) >> DD_YEAR_SHIFT;
 		days = year * 365;
 		days += year / 4 + 1;	/* add in leap days */
+		/*
+		 * XXX - works in 2000, but won't work in 2100.
+		 */
 		if ((year & 0x03) == 0)
 			days--;		/* if year is a leap year */
 		months = year & 0x03 ? regyear : leapyear;
@@ -270,91 +301,86 @@ smb_dos2unixtime(u_int dd, u_int dt, u_int dh, int tzoff,
 
 static int
 smb_fphelp(struct mbchain *mbp, struct smb_vc *vcp, struct smbnode *np,
-	int caseopt)
+	int caseopt, int *lenp)
 {
-	struct smbmount *smp= np->n_mount;
-	struct smbnode **npp = smp->sm_npstack;
+        struct smbnode  *npstack[SMBFS_MAXPATHCOMP]; 
+        struct smbnode  **npp = &npstack[0]; 
 	int i, error = 0;
 
-/*	simple_lock(&smp->sm_npslock);*/
 	i = 0;
 	while (np->n_parent) {
-		if (i++ == SMBFS_MAXPATHCOMP) {
-/*			simple_unlock(&smp->sm_npslock);*/
+		if (i++ == SMBFS_MAXPATHCOMP)
 			return ENAMETOOLONG;
-		}
 		*npp++ = np;
 		np = np->n_parent;
 	}
-/*	if (i == 0)
-		return smb_put_dmem(mbp, vcp, "\\", 2, caseopt);*/
 	while (i--) {
 		np = *--npp;
-#ifdef APPLE
 		if (SMB_UNICODE_STRINGS(vcp))
 			error = mb_put_uint16le(mbp, '\\');
 		else
 			error = mb_put_uint8(mbp, '\\');
-#else
-		error = mb_put_uint8(mbp, '\\');
-#endif
+                if (!error && lenp)
+                        *lenp += SMB_UNICODE_STRINGS(vcp) ? 2 : 1;
 		if (error)
 			break;
-		error = smb_put_dmem(mbp, vcp, np->n_name, np->n_nmlen, caseopt);
+		error = smb_put_dmem(mbp, vcp, np->n_name, np->n_nmlen,
+				     caseopt, lenp);
 		if (error)
 			break;
 	}
-/*	simple_unlock(&smp->sm_npslock);*/
 	return error;
 }
 
 int
 smbfs_fullpath(struct mbchain *mbp, struct smb_vc *vcp, struct smbnode *dnp,
-	const char *name, int nmlen)
+	const char *name, int *lenp, u_int8_t sep)
 {
 	int caseopt = SMB_CS_NONE;
-	int error;
+	int error, len = 0;
 
-#ifdef APPLE
+        if (lenp) {
+                len = *lenp;
+                *lenp = 0;
+        }
 	if (SMB_UNICODE_STRINGS(vcp)) {
 		error = mb_put_padbyte(mbp);
 		if (error)
 			return error;
 	}
-#endif
 	if (SMB_DIALECT(vcp) < SMB_DIALECT_LANMAN1_0)
 		caseopt |= SMB_CS_UPPER;
 	if (dnp != NULL) {
-		error = smb_fphelp(mbp, vcp, dnp, caseopt);
+		error = smb_fphelp(mbp, vcp, dnp, caseopt, lenp);
 		if (error)
 			return error;
 		if (dnp->n_ino == 2 && !name)
 			name = ""; /* to get one backslash below */
 	}
 	if (name) {
-#ifdef APPLE
 		if (SMB_UNICODE_STRINGS(vcp))
-			error = mb_put_uint16le(mbp, '\\');
+			error = mb_put_uint16le(mbp, sep);
 		else
-			error = mb_put_uint8(mbp, '\\');
-#else
-		error = mb_put_uint8(mbp, '\\');
-#endif
+			error = mb_put_uint8(mbp, sep);
+                if (!error && lenp)
+                        *lenp += SMB_UNICODE_STRINGS(vcp) ? 2 : 1;
 		if (error)
 			return error;
-		error = smb_put_dmem(mbp, vcp, name, nmlen, caseopt);
+		error = smb_put_dmem(mbp, vcp, name, len, caseopt, lenp);
 		if (error)
 			return error;
 	}
 	error = mb_put_uint8(mbp, 0);
-#ifdef APPLE
-	if (SMB_UNICODE_STRINGS(vcp) && error == 0)
+        if (!error && lenp)
+                *lenp++;
+	if (SMB_UNICODE_STRINGS(vcp) && error == 0) {
 		error = mb_put_uint8(mbp, 0);
-#endif
+                if (!error && lenp)
+                        *lenp++;
+	}
 	return error;
 }
 
-#ifdef APPLE
 void
 smbfs_fname_tolocal(struct smbfs_fctx *ctx)
 {
@@ -394,16 +420,3 @@ smbfs_fname_tolocal(struct smbfs_fctx *ctx)
 	return;
 }
 
-#else
-
-void
-smbfs_fname_tolocal(struct smb_vc *vcp, char *name, int nmlen, int caseopt)
-{
-/*	if (caseopt & SMB_CS_UPPER)
-		iconv_convmem(vcp->vc_toupper, name, name, nmlen);
-	else if (caseopt & SMB_CS_LOWER)
-		iconv_convmem(vcp->vc_tolower, name, name, nmlen);*/
-	if (vcp->vc_tolocal)
-		iconv_convmem(vcp->vc_tolocal, name, name, nmlen);
-}
-#endif
