@@ -9,7 +9,7 @@
 #include <sys/malloc.h>
 #include <sys/smb_apple.h>
 #include <sys/utfconv.h>
-#include <sys/iconv.h>
+#include <sys/smb_iconv.h>
 
 #include "iconv_converter_if.h"
 
@@ -43,7 +43,7 @@ iconv_utf8_open(struct iconv_converter_class *dcp,
 	#pragma unused(cspf)
 	struct iconv_utf8 *dp;
 
-	dp = (struct iconv_utf8 *)kobj_create((struct kobj_class*)dcp, M_ICONV, M_WAITOK);
+	dp = (struct iconv_utf8 *)kobj_create((struct kobj_class*)dcp, M_ICONV);
 	if (strcmp(csp->cp_to, "utf-8") == 0) {
 		dp->d_type = UTF8_ENCODE;
 		dp->d_flags = UTF_DECOMPOSED | UTF_NO_NULL_TERM;
